/* 
 * CS4451B - Project 2: OpenGL Animation
 * by Jacob Bolton - gte189u
 * 
 * Objectives:
 *   |
 *   +- Camera motion - The camera revolves (x,z) around the origin using a sine wave for y.
 *   |
 *   +- Object animation - There are several animations occuring. See the following functions:
 *   |                     draw_cubeRotate, draw_teapotSmoking, draw_bouncingBall, 
 *   |                     draw_clouds, draw_waves
 *   |
 *   +- Object instancing - draw_trees and draw_clouds use command lists. 
 *   |                      See init_trees & init_clouds.
 *   |
 *   +- Object contact - The trees are all located against the ground plane.  The teapot is as
 *   |                   close as I could get without actually knowing the dimensions.
 *   |
 *   +- Duration - The animation is a sequence of 200 looped frames.  The animation is played
 *   |             through once to assess the machine's ideal framerate.  The following frames
 *   |             are attempted to be played at a rate based on the assesment.
 *   |
 *   +- Lighting and shading - A light source was provided and unaltered.  Shiny shading was used
 *   |                         in draw_teapotSmoking, draw_bouncingBall, draw_sun
 *   |
 *   +- Perspective projection - Perspective projection was provided and unaltered.
 *   |
 *   +- Double-buffering - Double-buffering was provided and unaltered.
 *   |
 *   +- Hidden surfaces -  - z-buffering was provided and unaltered.
 *
 * Based on provided code:
 * Example animation using GLUT routines.
 * Greg Turk, April 1998
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <glut.h>

/* number of frames in the animation */
#define FRAMES 200

#ifndef PI
//in case someone forgot pi somewhere
#define PI 3.14159265358979
#endif

/* light position */
GLfloat  lightPos[] = { 1.0, 1.0, 1.0, 0.0 };

/* light contribution */
GLfloat  ambientLight[] = { (float)0.3, (float)0.3, (float)0.3, 0.0 };
GLfloat  diffuseLight[] = { (float)0.7, (float)0.7, (float)0.7, 0.0 };
GLfloat  specularLight[] = { (float)1.0, (float)1.0, (float)1.0, 0.0};

/* two examples of specular contributions */
GLfloat  shiny_surface[] =  { 1.0, 1.0, 1.0, 0.0 };
GLfloat  diffuse_surface[] =  { 0.0, 0.0, 0.0, 0.0 };

/* positions for trees */
int TREE_LIST = 1;
float rx[] = {50.0f, -50.0f, -50.0f, 50.0f, 10.0f};
float rz[] = {50.0f, -50.0f, 50.0f, -50.0f, 10.0f};

/* overall position for cloud group */
int CLOUD_LIST = 2;
float cx = 0.0;
float cz = 0.0;
float cr = 0.0;

/* current frame of the animation */
int currentFrame = -1;

/* milliseconds per frame - calculated in display */
int ms_f = 0;

/* draws a circular plane for the ground */
void draw_ground() {
	/* draw the ground plane (i like circular planes) */
	GLUquadricObj* quadObj;
	glMaterialfv(GL_FRONT, GL_SPECULAR, diffuse_surface);
	glColor3f (0.0, 0.75, 0.0);	//green
	glPushMatrix();
	glRotatef(90.0, 1.0, 0.0, 0.0);	//rotate to face the correct way
	quadObj = gluNewQuadric();
	gluDisk(quadObj, 0.0, 1000.0, 8, 8);
	gluDeleteQuadric(quadObj);
	glPopMatrix();
}

/* draws the camera based on som parameters
 * 
 * i - current frame
 * radius - radius for revolution around origin
 * maxHeigth - maximum height for vertical sine wave motion
 * dip - how far below the ground plane the camera will go
 */
void draw_camera(int i, double radius, double maxHeight, double dip) {
	double cam_x, cam_y, cam_z;
	//revolve camera around origin at radius
	cam_x = cos(((double)i / (double)FRAMES)*2.0*PI) * radius;
	cam_z = sin(((double)i / (double)FRAMES)*2.0*PI) * radius;
	cam_y = (-(sin(((double)i / (double)FRAMES)*PI)) 
		+ (double)0.7) * maxHeight + (maxHeight - dip);
	/* specify the camera position (from, at, up) */
	gluLookAt (cam_x, cam_y, cam_z,   0.0, 0.0, 0.0,  0.0, 1.0, 0.0);

}

/* draws a rotating cube given i, the current frame */
void draw_cubeRotate(int i){
  /* no specular contribution */
  glMaterialfv(GL_FRONT, GL_SPECULAR, diffuse_surface);
  glColor3f (0.4f, 0.0f, 0.0f);	//dark red
  glPushMatrix();
  glRotatef ((float) (-i*i), 0.0, 1.0, 0.0);	//rotate based on current frame
  glTranslatef(0.0f, -5.0f, 0.0f);	//move down a bit
  glutSolidCube (5.0);
  glPopMatrix();
}

/* draws a sequence of smoke puffs
 *
 * i - current frame
 * puffs - limit of puffs to produce
 * height - how high the puffs can go before dissipating
 * expanse - how wide the puffs get as they rise
 */
void draw_smoke(int i, int puffs, float height, float expanse) {
	int c, fpp;
	float hpp, dy, y_pos, expansion;
	fpp = FRAMES / puffs;
	hpp = height / (float)puffs;
	dy = height / (float)FRAMES;
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, diffuse_surface);
	for(c = 0; c < (i / fpp); c++) {
		y_pos = ((float)i * dy) - (c * hpp);
		expansion = 1.0f + (y_pos * y_pos) / (height * height) * (expanse - 1.0f);
		glColor4f (0.9f, 0.9f,0.9f, 0.7f - (y_pos / height) * 0.7);	//transparent grey
		glPushMatrix();
		glTranslatef(0.0, y_pos, 0.0);
		glScalef(expansion, 1.0, expansion);
		glutSolidSphere (0.4, 16, 16);
		glPopMatrix();
	}
}

/* draws a glut teapot with smoke puffs */
void draw_teapotSmoking(int i){
	glMaterialfv(GL_FRONT, GL_SPECULAR, shiny_surface);
	glMateriali(GL_FRONT,GL_SHININESS,128);
	glColor3f (0.6f, 0.6f, 0.6f);	//shiny gray
	glutSolidTeapot(2.0);
	glPushMatrix();
	glTranslatef(3.1f, 0.8f, 0.0);
	draw_smoke(i, 10, 4.0, 5.0);
	glPopMatrix();
}

/* draws a branch for a tree */
void draw_branch() {
	GLUquadricObj* quadObj;
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, diffuse_surface);
	glColor3f(0.3f, 0.3f, 0.0f);	//brown
	quadObj = gluNewQuadric();
	gluCylinder(quadObj, 1.0, 0.5, 10.0,16, 4);
	glPushMatrix();
	glColor3f(0.0f, 0.5f, 0.0f);	//green
	glTranslatef(0.0f, 0.0f, 11.0f);
	glutSolidSphere (3, 8, 8);
	glPopMatrix();
	gluDeleteQuadric(quadObj);
}

/* initializes a GL command list for making a tree */
void init_treeList() {
	glNewList(TREE_LIST, GL_COMPILE);
	GLUquadricObj* quadObj;
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, diffuse_surface);
	glColor3f(0.3f, 0.3f, 0.0f);	//brown
	quadObj = gluNewQuadric();
	glPushMatrix();
	glRotatef(90, -1.0f, 0.0f, 0.0f);
	gluCylinder(quadObj, 2.0, 1.0, 25.0,8, 4);	//trunk
	glTranslatef(0.0f, 0.0f, 25.0f);
	gluCylinder(quadObj, 1.0, 0.0, 5.0,8, 4);	//tip
	glPopMatrix();
	gluDeleteQuadric(quadObj);
	//add some levels of branches
	for(int j = 0; j < 5; j++) {
	for(int i = 0; i < (8 - j); i++) {
		float scale = 1.0f - 0.8f * ((float)j / 5.0f);
		glPushMatrix();
		glRotatef(360.0f / (8.0f - (float)j) *(float)i + ((float)j * 22.5f), 0.0f, 1.0f, 0.0f);
		glTranslatef(0.0f, 5.0f + (5.0f * (float)j), 0.1f);
		//random branch rotation looks better
		glRotatef((((float)rand() / (float)RAND_MAX) - 0.5f) * 30.0f, 1.0f, 0.0f, 0.0f);
		glScalef(scale, scale, scale);
		draw_branch();
		glPopMatrix();
	}
	}
	glEndList();
}

//probably should move this to the top and use it in more functions one of these days
#define randf (((float)rand() - ((float)RAND_MAX / 2.0f)) / (float)RAND_MAX)

/* initializes a GL command list for making a set of clouds */
void init_cloudList() {
	glNewList(CLOUD_LIST, GL_COMPILE);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, diffuse_surface);
	for(int i = 0; i < 20; i++) {
		glColor4f(1.0f, 1.0f, 1.0f, (randf + 1.0f) / 2.0f);	//transparent white
		glPushMatrix();
		glTranslatef(randf * 50.0f, 0.0f, randf * 50.0f);
		glScalef(randf * 5.0f, randf * 2.0f, randf * 5.0f);
		glutSolidSphere (3.0f, 8, 8);
		glPopMatrix();
	}
	glEndList();
}

/* draws a bouncing ball
 *
 * i - the current frame
 * (x1,z1) - starting location on the ground (y = 0) plane (y+ is up)
 * (x2,z2) - ending location on the ground (y = 0) plane (y+ is up)
 * maxHeight - height of the ball's arc
 * bounces - number of bounces to perform
 */
void draw_bouncingBall(int i, float x1, float z1, float x2, float z2, double maxHeight, 
					   float bounces) {
	float dx, dz,expand;
	double height, dt, compress;
	dx = (x2 - x1) / (float)FRAMES;
	dz = (z2 - z1) / (float)FRAMES;
	dt = ((double)bounces * (double)PI) / (double)FRAMES;
	height = sin(dt * (double)i);
	height = (height + 1.0) / 2;
	height = height * maxHeight;
	//vertical compression
	if(height < 3.0) {
		compress = 0.5 + height / 6.0;
	} else {
		compress = 1.0 + ((double)rand() / (double)RAND_MAX) * 0.05;
	}
	//horizontal expansion
	if(height < 1.0) {
		expand = 1.25 - height / 12.0;
	} else {
		expand = 1.0 - ((double)rand() / (double)RAND_MAX) * 0.05;
	}
	glMaterialfv(GL_FRONT, GL_SPECULAR, shiny_surface);
	glMateriali(GL_FRONT,GL_SHININESS,128);
	glColor3f (1.0f, 0.0f, 0.0f);	//red
	glPushMatrix();
	glTranslatef(x1 + (float)i * dx, 
		(1.5f * (float)compress) + (float)height, z1 + (float)i * dz);
	glScalef(expand, compress, expand);
	glutSolidSphere (3.0f, 32, 32);
	glPopMatrix();
}

/* draws the clouds using a slightly randomized position and rotation */
void draw_clouds() {
	//maybe change the position and rotation a bit
	if(randf > 0){
		cx = cx + randf*2.0f;
		cz = cz + randf*2.0f;
		cr = cr + randf*5.0f;
		if(cx > 20.0f){
			cx = 10.0f;
		} else if(cx < -10.f) {
			cx = -20.0f;
		}
		if(cz > 20.0f){
			cz = 10.0f;
		} else if(cz < -10.f) {
			cz = -20.0f;
		}
		if(cr >= 360.0f){
			cr = 0.0f;
		} else if(cr <= -360.0f){
			cr = 0.0f;
		}
	}
	glPushMatrix();
	glTranslatef(cx, 20.0f, cz);
	glRotatef(cr, 0, 1, 0);
	glCallList(CLOUD_LIST);
	glPopMatrix();
}

/* draws a sun (no emission) */
void draw_sun() {
	glMaterialfv(GL_FRONT, GL_SPECULAR, shiny_surface);
	glMateriali(GL_FRONT,GL_SHININESS,128);
	glPushMatrix();
	glColor3f (1.0f, 1.0f, 0.0f);	//yellow
	glTranslatef(-80.0f, 50.0f, -80.0f);
	glutSolidSphere (20.0f, 16, 16);
	glPopMatrix();
}

/* draws a field of waves
 *
 * i - the current frame
 * width, depth - the dimension of the field
 * maxHeight - the height of the waves' crests and troughs
 * waves - the number of waves
 */
void draw_waves(int i, int width, int depth, double maxHeight, double waves) {
	double height = maxHeight / 2.0;
	double dt = waves * 2.0 * (double)PI / (double)FRAMES;
	double dz = waves * 2.0 * (double)PI / (double)depth;
	double dx = waves * 2.0 * (double)PI / (double)width;
	glMaterialfv(GL_FRONT, GL_SPECULAR, diffuse_surface);
	glColor3f (0.5f, 0.0f, 0.2f);	//darkER red
	glPushMatrix();
	glTranslatef(-(float)width / 2.0f, 0.0f, -(float)depth / 2.0f);
	for(int x = 0; x < width; x++){
		for(int z = 0; z < depth; z++){
			double y1 = sin(z*dz + i * dt) * height;
			double y3 = cos(x*dx + i * dt) * height;
			double y2 = sin((z+1)*dz + i * dt) * height;
			double y4 = cos((x+1)*dx+ i * dt) * height;
			double nx = (y4 - y3);
			double ny = 1;
			double nz = (-(y1+y3)+(y2+y4));
			glBegin(GL_POLYGON);
			glNormal3f(nx,ny,nz);
			glVertex3f(x,y1 + y3,z);
			glVertex3f(x + 1,y1 + y4,z);
			glVertex3f(x + 1,y2 + y4,z + 1);
			glEnd();
			nx = (y4 - y3);
			ny = 1;
			nz = ((y2 + y4)-(y1 +y3)) - (y4 - y3);
			glBegin(GL_POLYGON);
			glNormal3f(nx,ny,nz);
			glVertex3f(x + 1,y2 + y4,z + 1);
			glVertex3f(x,y2 +y3,z + 1);
			glVertex3f(x,y1 + y3,z);
			glEnd();
		}
	}
	glPopMatrix();
}

/******************************************************************************
Draw the animation scene.
******************************************************************************/

void draw_scene(int i)
{
	glPushMatrix();
	draw_camera(i, 30, 10, 7.2);
	draw_ground();
	draw_cubeRotate(i);
	glPushMatrix();
	glTranslatef(0.0f, 1.5f, 0.0f);
	glRotatef(45, 0,1,0);
	draw_teapotSmoking(i);
	glPopMatrix();
	for(int t = 0; t < 5; t++){
		glPushMatrix();
		glTranslatef(rx[t], 0.0f, rz[t]);
		glScalef(0.5f, 0.5f, 0.5f);
		glCallList(TREE_LIST);
		glPopMatrix();
	}
	draw_bouncingBall(i, -5.0f, -10.0f, -5.0f, 10.0f, 10.0f, 10.0f);
	draw_clouds();
	draw_sun();
	glPushMatrix();
	glTranslatef(0.0f, -20.0f, 0.0f);
	glRotatef(45, 0.0f, 1.0f, 0.0f);
	glScalef(5.0f, 1.0f, 5.0f);
	draw_waves(i, 24, 24, 5.0, 3);
	glPopMatrix();
	glPopMatrix();
}

/* Increments the current frame by frames and draws the new frame */
void displayFrame(int frames)
{  
	int i;
	currentFrame += frames;
	i = currentFrame = currentFrame % FRAMES;
   
    /* clear the color and z-buffer */
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    /* set up the perspective transformation */
    glMatrixMode (GL_PROJECTION);
    glLoadIdentity();
    gluPerspective (60.0, 1.0, 1.0, 1000.0);

    /* from now on we're affecting the modeling matrix */
    glMatrixMode(GL_MODELVIEW);

    /* draw our scene */
    draw_scene (i);

    /* swap that performs double-buffering */
    glutSwapBuffers();

	//if the milli-frame delay is known, use it
	if(ms_f > 0){
		glutTimerFunc((unsigned int)ms_f, displayFrame, 1);
	}
}

/* display function for glut's redraw and such */
void display()
{
	//just starting? must do framrate calculation
	if(currentFrame == -1){
		//grab current time and zoom through the animation
		int start = glutGet(GLUT_ELAPSED_TIME);
		for(int i = 0; i < FRAMES; i++)
			displayFrame(1);
		ms_f = (glutGet(GLUT_ELAPSED_TIME) - start) / FRAMES + 1;
#ifdef DEBUG
		printf("ms_f:%d", ms_f);
#endif
		//don't really need/want more than 50fps, that's only 4s
		if(ms_f < 20) {
			ms_f = 20;
		}
	//not just starting? then just redraw the current frame
	} else {
		displayFrame(0);
	}
}


/******************************************************************************
Initilize various graphics things such as surface material and lights.
******************************************************************************/

void init_graphics()
{
  /* enable depth buffering */
  glEnable(GL_DEPTH_TEST);   

  /* surface shading info */
  glFrontFace(GL_CCW); 
  
  //Yay for alpha channel!!!
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  
  glShadeModel (GL_SMOOTH);

  /* these lines allow the diffuse and ambient color to be set by */
  /* the gtColor3f() commands that proceed an object */

  glColorMaterial (GL_FRONT, GL_AMBIENT_AND_DIFFUSE); 
  glEnable (GL_COLOR_MATERIAL);

  /* turn on lighting */
  glEnable (GL_LIGHTING);


  /* create intensity contributions of the light */
  glLightfv(GL_LIGHT0, GL_AMBIENT, ambientLight);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuseLight);
  glLightfv(GL_LIGHT0, GL_SPECULAR, specularLight);

  /* specify the light position */
  glLightfv(GL_LIGHT0,GL_POSITION,lightPos);
  
  /* turn on light zero */
  glEnable(GL_LIGHT0);
  
  /* select clearing color */ 
  glClearColor (0.2f, 0.2f, 0.8f, 1.0f);

  //initialize GL command lists
  init_treeList();
  init_cloudList();
}


/******************************************************************************
Process a keyboard action.  In particular, exit the program when an
"escape" is pressed in the window.
******************************************************************************/

void keyboard(unsigned char key, int x, int y)
{
  switch (key) {
    case 27:
      exit(0);
      break;
  }
}




/******************************************************************************
Main routine.
******************************************************************************/

int main(int argc, char** argv)
{
  glutInit(&argc, argv);
  glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutInitWindowSize (250, 250); 
  glutInitWindowPosition (100, 100);
  glutCreateWindow (argv[0]);

  init_graphics();

  glutDisplayFunc (display);
  //animate based on time elapsed, not only on redraw requests
  //this works only so-so, i didn't feel like figuring out a better way to do it with glut
  glutTimerFunc((unsigned int)100, displayFrame, 1);
  glutKeyboardFunc (keyboard);
  glutMainLoop(); 
  return 0;
}

